using System;
using System.Collections.Generic;
using System.Text;

namespace FM
{
	/// <summary>
	/// Trieda uchovva informcie o konkrtnej vlastnosti. Interpretuje informcie o svojej kardinalite 
	/// (vie zisti, i ide o povinn alebo voliten vlastnos), implementuje metdu na vpis informci
	/// o sebe ako aj metdu na validciu vlastnosti.
	/// </summary>
    [Serializable]
    public class DFeature : CardinalityHolder
    {

        private DFeature lastAddedFeature = null;

		/// <summary>
		/// Jeden z kontruktorov. Prid vlastnos do stromu a nastav jej identifiktor (zabezpeuj to 
		/// nadraden triedy). Nastav kardinalitu skupiny na 1..1 (povinn vlastnos).
		/// </summary>
		/// <param name="parent">Odkaz na predka, pod ktorho bude zaraden nov uzol. V prpade, e ide o kore stromu, predva sa hodnota null, no objekt tejto triedy by nemal by nikdy koreom.</param>
		/// <param name="ID">Identifiktor uzla. Uniktnos aj konverziu nepovolench znakov zabezpeuje nadraden trieda.</param>
		public DFeature(TreeNode parent, String ID)
			:
			base(parent, ID)
		{
			this.Cardinality.SetCardinality(1, 1);
			this.AiItems.ImportAiItems(Diagram.NodeAiTemplates.FeatureAiItems);
		}

		/// <summary>
		/// Jeden z kontruktorov. Prid vlastnos do stromu a nastav jej identifiktor (zabezpeuj to 
		/// nadraden triedy). Nastav minimlnu kardinalitu vlastnosti na minCardinality a maximlnu automaticky 
		/// na nekoneno.
		/// </summary>
		/// <param name="parent">Odkaz na predka, pod ktorho bude zaraden nov uzol. V prpade, e ide o kore stromu, predva sa hodnota null, no objekt tejto triedy by nemal by nikdy koreom.</param>
		/// <param name="ID">Identifiktor uzla. Uniktnos aj konverziu nepovolench znakov zabezpeuje nadraden trieda.</param>
		/// <param name="minCardinality">Minimlna kardinalita vlastnosti. Maximlna bude automaticky nastaven na nekoneno.</param>
		public DFeature(TreeNode parent, String ID, int minCardinality)
			:
			base(parent, ID)
		{
			this.Cardinality.SetCardinality(minCardinality);
			this.AiItems.ImportAiItems(Diagram.NodeAiTemplates.FeatureAiItems);
		}

		/// <summary>
		/// Jeden z kontruktorov. Prid vlastnos do stromu a nastav jej identifiktor (zabezpeuj to 
		/// nadraden triedy). Nastav kardinalitu vlastnosti na minCardinality..maxCardinality.
		/// </summary>
		/// <param name="parent">Odkaz na predka, pod ktorho bude zaraden nov uzol. V prpade, e ide o kore stromu, predva sa hodnota null, no objekt tejto triedy by nemal by nikdy koreom.</param>
		/// <param name="ID">Identifiktor uzla. Uniktnos aj konverziu nepovolench znakov zabezpeuje nadraden trieda.</param>
		/// <param name="minCardinality">Minimlna kardinalita vlastnosti.</param>
		/// <param name="maxCardinality">Maximlna kardinalita vlastnosti.</param>
		public DFeature(TreeNode parent, String ID, int minCardinality, int maxCardinality)
			:
			base(parent, ID)
		{
			this.Cardinality.SetCardinality(minCardinality, maxCardinality);
			this.AiItems.ImportAiItems(Diagram.NodeAiTemplates.FeatureAiItems);
		}

        /// <summary>
        /// Naimportuje AI ablnu.
        /// </summary>
        public void ImportAiTemplate()
        {
            this.AiItems.ImportAiItems(Diagram.NodeAiTemplates.FeatureAiItems);
        }

		/// <summary>
		/// Informcia o tom, i je vlastnos vybran (pri konfigurci).
		/// </summary>
		private bool selected;
		/// <summary>
		/// Zist, i je vlastnos vybran (m vznam len pri konfigurci, inak vdy vrti nepravdu).
		/// Resp. nastav vlastnos ako vybran (len v prpade, e ide o konfigurciu).
		/// </summary>
		public bool Selected
		{
			get
			{
				if (Diagram.IsInstance)
				{
					return selected;
				}
				else
				{
					// Patrilo by sa hodit nejaku vynimku asi, ale zatial len takto:
					return false;
				}
			}
			set
			{
				if (Diagram.IsInstance)
				{
					selected = value;
				}
				else
				{
					// Patrilo by sa hodit nejaku vynimku asi, ale zatial len takto:
					return;
				}
			}
		}

        /// <summary>
        /// Vracia odkaz na posledne pridan feature. Vyuva sa pri rieen viacnsobnch features.
        /// </summary>
        public DFeature LastAddedFeature
        {
            get
            {
                return this.lastAddedFeature;
            }
        }

        /// <summary>
        /// Nastav vlastnost na vybran (len v prpade, e ide o konfigurciu). Zrove oetruje viacnsobn features.
        /// </summary>
        /// <returns>Oznami, ci je potrebny refresh ovladacieho prvku TreeView (ci bol pridany feature)</returns>
		public bool Select()
		{
			if (Diagram.IsInstance)
			{
                bool needRefresh = false;
                bool noMoreUnselected = true;
                int cntFeatures = 1;
                foreach (TreeNode neighbour in this.GetNeighbours())
                {
                    if (neighbour.GetType() != typeof(DFeature))
                    {
                        continue;
                    }

                    DFeature neighbourFeature = (DFeature)neighbour;
                    if (neighbourFeature.Name == this.Name)
                    {
                        cntFeatures++;
                        if (neighbourFeature.Selected == false)
                        {
                            noMoreUnselected = false;
                        }
                    }
                }

                if (noMoreUnselected && (cntFeatures < this.Cardinality.Max || this.Cardinality.Infinite))
                {
                    this.lastAddedFeature = (DFeature)this.CloneTree();
                    this.Parent.AddChildAfter(this.lastAddedFeature, this);
                    needRefresh = true;
                }

                selected = true;

                return needRefresh;
            }
			else
			{
				// Patrilo by sa hodit nejaku vynimku asi, ale zatial len takto:
				return false;
			}
		}

		/// <summary>
		/// Nastav vlastnos ako nevybran (len v prpade, e ide o konfigurciu).
		/// </summary>
		public void Deselect()
		{
			if (Diagram.IsInstance)
			{
				selected = false;
			}
			else
			{
				// Patrilo by sa hodit nejaku vynimku asi, ale zatial len takto:
				return;
			}
		}

		/// <summary>
		/// Prepne nastavenie vlastnosti (ak bola vybran, bude nevybrana a naopak). Akcia sa vykon
		/// len v prpade, e ide o konfigurciu.
		/// </summary>
		public void ToggleSelect()
		{
			if (Diagram.IsInstance)
			{
				selected = !selected;
			}
			else
			{
				// Patrilo by sa hodit nejaku vynimku asi, ale zatial len takto:
				return;
			}
		}

		/// <summary>
		/// Zist, i ide o povinn vlastnos (kardinalita 1..1), resp. ju nastav ako povinn.
		/// </summary>
		public bool Mandatory
		{
			get
			{
//                if (this.Cardinality.Min == 1 && this.Cardinality.Max == 1)
				if (this.Cardinality.Min > 0) // TT: nutne pre * kardinality - je povinna aspon jedna
                {
					return true;
				}
				else
				{
					return false;
				}
			}
			set
			{
				if (value == true)
				{
					this.Cardinality.SetCardinality(1, 1);
				}
			}
		}

		/// <summary>
		/// Zist, i ide o voliten vlastnos (kardinalita 0..1), resp. ju nastav ako voliten.
		/// </summary>
		public bool Optional
		{
			get
			{
//                if (this.Cardinality.Min == 0 && this.Cardinality.Max == 1)
				if (this.Cardinality.Min == 0) // TT: nutne pre * kardinality, nie je povinna ziadna
                {
					return true;
				}
				else
				{
					return false;
				}
			}
			set
			{
				if (value == true)
				{
					this.Cardinality.SetCardinality(0, 1);
				}
			}
		}

		/// <summary>
		/// Preaen metda, sluiaca na vpis informci o vlastnosti (vype meno, kardinalitu a prpadne aj i ide o 
		/// povinn alebo voliten vlastnos).
		/// </summary>
		/// <returns>Reazec s informciami o vlastnosti, ktor sa d vypsa do textovho prvku formulra alebo do konzoly.</returns>
		override protected String Dump()
		{
			String ret = "";
            ret += base.Dump();
            ret += "Type: Feature\r\n";
			ret += "Cardinality: " + this.Cardinality.Cardinality + " ";
			if (this.Mandatory == true)
			{
				ret += "(mandatory feature)";
			}
			if (this.Optional == true)
			{
				ret += "(optional feature)";
			}
			ret += "\r\n";

            if (this.Diagram.IsInstance)
            {
				if (this.Selected)
				{
					ret += "Selected\r\n";
				}
				else
				{
					ret += "Deselected\r\n";
				}

                if (this.Validate())
                {
                    ret += "Cardinality validity: valid\r\n";
                }
                else
                {
                    ret += "Cardinality validity: invalid\r\n";
                }
            }

			return ret;
		}

		/// <summary>
		/// Preaen metda zisuje, i ide o platn (validn) vlastnos. Uren hlavne pre budce pouitie.
		/// V tomto momente vracia vdy pravdu.
		/// </summary>
		/// <returns>Momentlne vdy vracia pravdu.</returns>
		override public bool Validate()
		{
			if (Unavailable == false && this.Diagram.IsInstance)
			{
				int numSelected = 0;

				foreach (TreeNode neighbour in this.GetNeighbours())
				{
					if (neighbour.GetType() != typeof(DFeature))
					{
						continue;
					}

					DFeature neighbourFeature = (DFeature)neighbour;
					if (neighbourFeature.Name == this.Name && neighbourFeature.Selected)
					{
						numSelected++;
					}
				}

				if (this.Selected)
				{
					numSelected++;
				}

				if (numSelected >= this.Cardinality.Min && (numSelected <= this.Cardinality.Max || this.Cardinality.Infinite))
				{
					return true;
				}
				else
				{
					if (ReadOnly)
					{
						Diagram.CanBeValid = false;
					}
					return false;
				}
			}
			else
			{
				return true;
			}
		}

        /// <summary>
        /// Kopruje Feature.
        /// </summary>
        /// <returns>Vrti kpiu Feature.</returns>
        public override TreeNode Clone()
        {
            DFeature copy = (DFeature)base.Clone();
            copy.selected = this.Selected;
            return (TreeNode)copy;
        }

        /// <summary>
        /// Zist, i je feature mon meni vzhadom an to, i je jej as viazania vi ako as intancializcie.
        /// </summary>
        public bool ReadOnly
        {
            get
            {
                return this.NodeBindingTime.Before(this.Diagram.NodeBindingTime);
            }
        }

		/// <summary>
		/// Zist, i je feature mon meni vzhadom an to, i je jej as viazania men ako as intancializcie.
		/// </summary>
		public bool Unavailable
		{
			get
			{
				return this.Diagram.NodeBindingTime.Before(this.NodeBindingTime);
			}
		}

        /// <summary>
        /// Zhodnot, i treba oznai Feature ako selected pri vytvaran intancie.
        /// Ak treba, klonuje samho seba (cel podstrom) poda hodnoty "Cardinality.min".
        /// </summary>
        public override void ResolveFeatureSelection()
        {
            // najprv prejdi cely podstrom
            base.ResolveFeatureSelection();

            // potom sa zaoberaj samym sebou
            if (Unavailable == false && ReadOnly == false && Mandatory) {
                this.selected = true;
                if (Cardinality.Min > 1) {
                    // vytvor tolko kopii, kolko je treba
                    TreeNode parent = this.Parent;

                    int i;
                    DFeature copy = null;
                    DFeature LastCopy = this;

                    for (i = 1; i < Cardinality.Min; i++) {
                        copy = (DFeature)this.CloneTree();
                        parent.AddChildAfter(copy, LastCopy);
                        LastCopy = copy;
                    }

                    if (this.Cardinality.Infinite)
                    {
                        copy = (DFeature)this.CloneTree();
                        copy.selected = false;
                        parent.AddChildAfter(copy, LastCopy);
                    }
                    else
                    {
                        for (int a = i; a < this.Cardinality.Max; a++)
                        {
                            copy = (DFeature)this.CloneTree();
                            copy.selected = false;
                            parent.AddChildAfter(copy, LastCopy);
                            LastCopy = copy;
                        }
                    }
                }
            }
        }
    }
}
