using System;
using System.Collections.Generic;
using System.Text;

namespace FM
{
	/// <summary>
	/// Trieda uchovva informcie o skupine vlastnost. Interpretuje informcie o kardinalite (vie zisti,
	/// i ide o alternatvnu skupinu, alebo-skupiny, alebo je kardinalita uren explicitne). Implementuje metdu
	/// na vpis informci o sebe, ako aj metdu na svoju validciu.
	/// </summary>
    [Serializable]
    public class DGroup : CardinalityHolder

    {
		/// <summary>
		/// Skupina nema meno, namiesto neho sa generuje nadpis v tvare "G: [min...max]"
		/// </summary>
		override public String Name
		{
			get
			{
				return "G: [" + this.Cardinality.Cardinality + "]";				
			}
		}

        /// <summary>
        /// Jeden z kontruktorov. Prid skupinu do stromu a nastav jej identifiktor (zabezpeuj to 
        /// nadraden triedy). Nastav kardinalitu skupiny na 1..1 (alternatvna skupina)
        /// </summary>
        /// <param name="parent">Odkaz na predka, pod ktorho bude zaraden nov uzol. V prpade, e ide o kore stromu, predva sa hodnota null, no objekt tejto triedy by nemal by nikdy koreom.</param>
        /// <param name="ID">Identifiktor uzla. Uniktnos aj konverziu nepovolench znakov zabezpeuje nadraden trieda.</param>
        public DGroup(TreeNode parent, String ID)
            : base(parent, ID)
        {
            this.Cardinality.SetCardinality(1, 1);
        }

		/// <summary>
		/// Jeden z kontruktorov. Prid skupinu do stromu a nastav jej identifiktor (zabezpeuj to 
		/// nadraden triedy). Nastav minimlnu kardinalitu skupiny na minCardinality a maximlnu automaticky 
		/// na nekoneno. V prpade, e minCardinality bude 1, ide o alebo-skupinu (1..*).
		/// </summary>
		/// <param name="parent">Odkaz na predka, pod ktorho bude zaraden nov uzol. V prpade, e ide o kore stromu, predva sa hodnota null, no objekt tejto triedy by nemal by nikdy koreom.</param>
		/// <param name="ID">Identifiktor uzla. Uniktnos aj konverziu nepovolench znakov zabezpeuje nadraden trieda.</param>
		/// <param name="diagram">Odkaz na diagram, v ktorom sa uzol (a strom) nachdza.</param>
		/// <param name="minCardinality">Minimlna kardinalita skupiny. Maximlna bude automaticky nastaven na nekoneno.</param>
		public DGroup(TreeNode parent, String ID, Diagram diagram, int minCardinality)
			: base(parent, ID)
		{
			this.Cardinality.SetCardinality(minCardinality);
		}
		/// <summary>
		/// Jeden z kontruktorov. Prid skupinu do stromu a nastav jej identifiktor (zabezpeuj to 
		/// nadraden triedy). Nastav kardinalitu skupiny na minCardinality..maxCardinality.
		/// </summary>
		/// <param name="parent">Odkaz na predka, pod ktorho bude zaraden nov uzol. V prpade, e ide o kore stromu, predva sa hodnota null, no objekt tejto triedy by nemal by nikdy koreom.</param>
		/// <param name="ID">Identifiktor uzla. Uniktnos aj konverziu nepovolench znakov zabezpeuje nadraden trieda.</param>
		/// <param name="diagram">Odkaz na diagram, v ktorom sa uzol (a strom) nachdza.</param>
		/// <param name="minCardinality">Minimlna kardinalita skupiny.</param>
		/// <param name="maxCardinality">Maximlna kardinalita skupiny.</param>
		public DGroup(TreeNode parent, String ID, Diagram diagram, int minCardinality, int maxCardinality)
			:
			base(parent, ID)
		{
			this.Cardinality.SetCardinality(minCardinality, maxCardinality);
		}

		/// <summary>
		/// Zist, i ide o alternatvnu skupinu (kardinalita 1..1), resp. nastav skupinu ako alternatvnu v
		/// prpade, e sa hodnota bude true.
		/// </summary>
		public bool AlternativeGroup
		{
			get
			{
				if (this.Cardinality.Min == 1 && this.Cardinality.Max == 1)
				{
					return true;
				}
				else
				{
					return false;
				}
			}
			set
			{
				if (value == true)
				{
					this.Cardinality.SetCardinality(1, 1);
				}
			}
		}

		/// <summary>
		/// Zist, i ide o alebo-skupinu (kardinalita 1..k, kde k je poet potomkov, resp. 1..*), resp. nastav
		/// skupinu ako alebo-skupinu (kardinalita 1..*).
		/// </summary>
		public bool OrGroup
		{
			get
			{
				if (this.Cardinality.Min == 1 && (this.Cardinality.Max == this.CountChildren() || this.Cardinality.Infinite == true))
				{
					return true;
				}
				else
				{
					return false;
				}
			}
			set
			{
				if (value == true)
				{
					this.Cardinality.SetCardinality(1);
				}
			}
		}

		/// <summary>
		/// Preaen metda, sluiaca na vpis informci o skupine (vype meno, kardinalitu a prpadne aj i ide o 
		/// alternatvnu skupinu alebo alebo-skupinu).
		/// </summary>
		/// <returns>Reazec s informciami o skupine, ktor sa d vypsa do textovho prvku formulra alebo do konzoly.</returns>
        protected override String Dump()
		{
			String ret = "";
            ret += "Type: Group of features\r\n";
			ret += "Cardinality: " + this.Cardinality.Cardinality + " ";
			if (this.AlternativeGroup == true)
			{
				ret += "(alternative group)";
			}
			if (this.OrGroup == true)
			{
				ret += "(OR group)";
			}
			ret += "\r\n";

			if (this.Diagram.IsInstance)
			{
				if (this.Validate())
				{
					ret += "Cardinality validity: valid\r\n";
				}
				else
				{
					ret += "Cardinality validity: invalid\r\n";
				}
			}
			else
			{
				if (this.Validate() == false)
				{
					ret += "Warning: Group contains less than 2 features and therefore it is considered invalid.\r\n";
				}
			}

			return ret;
		}

		/// <summary>
		/// Preaen metda zisuje, i ide o platn (validn) skupinu. Uren hlavne pre budce pouitie.
		/// V tomto momente kontroluje, i poet vybranch potomkov zodpoved intervalu kardinality a i s
		/// platn vetci potomkovia.
		/// </summary>
		/// <returns>Pravda ak je skupina platn aj vetci jej potomkovia s platn, nepravda v opanom prpade.</returns>
        public override bool Validate()
        {
			if (this.Diagram.IsInstance)
			{
				int selectedCount = 0;
				foreach (DFeature child in Children)
				{
					if (child.Selected)
					{
						selectedCount++;
					}
				}

				if (selectedCount >= Cardinality.Min && (selectedCount <= Cardinality.Max || Cardinality.Infinite))
				{
                    return true;
				}
				else
				{
					return false;
				}
			}
			else
			{
				return (this.CountChildren() > 1);
			}
        }

        /// <summary>
        /// Poksi sa aby bolo v skupine oznaench aspo "Cardinality.Min" vlastnost.
        /// Treba dopracova, ak by sme chceli ma zaruen validnos intancie.
        /// V prpade, e by nebolo mon vytvori validn intanciu, mali by sme to oznmi pouvateovi.
        /// </summary>
        public override void ResolveFeatureSelection()
        {
            base.ResolveFeatureSelection();
            int selected = 0;
            // spota vybran vlastnosti
            foreach (DFeature item in Children)
                if (item.GetType() == typeof(DFeature) && item.Selected) selected++;

            int toselect = this.Cardinality.Min - selected;

            // ak nie je oznaen dostaton poet vlastnost, poksime sa oznai alie
            foreach (DFeature item in Children)
                if (toselect > 0 && item.GetType() == typeof(DFeature) && !item.Selected) {
                    item.Selected = true;
                    toselect--;
                }
            // nie je zaruen, e sa nm to podarilo
        }
    }
}
