using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;


namespace FM
{
	/// <summary>
	/// Trieda uchovva informcie o uzle v strome v rmci diagramu. Implementuje factory metdu na pridvanie
	/// potomkov ako aj metdy na zskanie priamych aj vzdialench potomkov urench relatvnou cestou. Definuje
	/// virtulnu metdu na validciu uzla.
	/// </summary>
    [Serializable]
    public class DNode : AiItemHolder
    {
        /// <summary>
        /// Docasny konstruktor. Je potrebne ho pred dokoncenim projektu zmazat!
        /// </summary>
        public DNode()
            : base(null, "")
        {
        }

        /// <summary>
		/// Kontruktor vytvor nov objekt typu DNode, priom ho automaticky zarad do stromu pod svojho
		/// predka (parent) a pridradi mu ID (nadraden trieda implementuje kontrolu na uniktnos a zabezpeuje
		/// konverziu nepovolench znakov).
		/// </summary>
		/// <param name="parent">Odkaz na predka, pod ktorho bude zaraden nov uzol. V prpade, e ide o kore stromu, predva sa hodnota null.</param>
		/// <param name="ID">Identifiktor uzla. Uniktnos aj konverziu nepovolench znakov zabezpeuje nadraden trieda.</param>
		public DNode(TreeNode parent, String ID)
			: base(parent, ID)
		{
//			this.diagram = diagram;
//			this.cardinality = new NodeCardinality();
        }
		/// <summary>
		/// Zska odkaz na diagram, v ktorom sa uzol nachdza.
		/// </summary>
		public Diagram Diagram
		{
			get
			{
                DNode current = (DNode)this.Parent;
                while ((current != null) && (current.GetType() != typeof(Diagram))) current = (DNode)current.Parent;
                if (current == null) return null;
                return (Diagram)current;
			}
		}

		/// <summary>
		/// Zska odkaz na projekt, v ktorom sa nachdza diagram, v ktorom sa uzol nachdza. 
		/// </summary>
		public Project Project
		{
			get
			{
                DNode current = (DNode)this.Parent;
                while ((current != null) && (current.GetType() != typeof(Project))) current = (DNode)current.Parent;
                if (current == null) return null;
                return (Project)current;
            }
        }

		/// <summary>
		/// Factory medda na pridanie uzla. Poda parametru type vytvor objekt zodpovedajcej triedy a prid ho 
		/// do stromu ako svojho potomka. Zatia me by typ jedno z "group" (prid sa objekt triedy DGroup) a 
		/// "feature", "reference", "link" (prid sa objekt triedy DFeature, o bude neskr zmenen pre posledn 
		/// dva zmienen typy).
		/// </summary>
		/// <param name="type">Typ objektu, ktor sa m vloi. Jedno z "group", "feature", "reference", "link" </param>
		/// <param name="ID">Identifiktor pre novho uzla.</param>
        public DNode AddNode(String type, String ID)
        {
            DNode newNode = null;
			switch (type)
			{
				case "group":
					newNode = new DGroup((TreeNode)this, ID);
					break;
				case "reference":
//                    DReference referenceChild = new DReference((TreeNode)this, ID);
//                    break;
				case "feature":
                    newNode = new DFeature((TreeNode)this, ID);
                    break;
                default:
//                    CardinalityHolder cardinalityHolderChild = new CardinalityHolder((TreeNode)this, ID);
					break;
			}

            return newNode;
        }

        /// <summary>
        /// Vrti odkaz na svojho bezprostrednho potomka s identifiktorom ID.
        /// </summary>
        /// <param name="ID">Identifiktor bezprostrednho potomka.</param>
        /// <returns>Odkaz na svojho bezprostrednho potomka s identifiktorom ID alebo null ak tak neexistuje.</returns>
        public DNode GetNode(String ID)
        {
            return (DNode)GetChild(ID);
        }

        /// <summary>
        /// Vrti odkaz na potomka s identifiktorom ID, ktory nasleduje za lastNode.
        /// </summary>
        /// <param name="ID"></param>
        /// <param name="lastNode"></param>
        /// <returns></returns>
        public DNode GetNode(String ID, DNode lastNode)
        {
            DNode found = (DNode)GetChild(ID, (AiItemHolder)lastNode);
            if (found == lastNode) return null;
            else return found;
        }

		/// <summary>
		/// Vrti odkaz na potomka (nemus s o bezprostrednho potomka) s cestou pathString (v bodkovom zpise) 
		/// relatvnou vzhadom na tento (this) uzol. Absoltna cesta sa zska spojenm cesty od korea tohto uzla 
		/// a uvedenej relatvnej cesty pathString. Metda sli hlavne na zjednoduenie a sprehadnenie zpisu ako aj
		/// na uahenie prce s vrazmi (expressions).
		/// </summary>
		/// <param name="pathString">Relatvna cesta k potomkovi vzhadom na tento uzol.</param>
		/// <returns>Odkaz na potomka, alebo null ak pre uveden relatvnu cestu neexistuje iadny potomok.</returns>
		public DNode GetNodeFromPath(String pathString)
		{
            return GetNodeFromPath(pathString, null);
        }

        /// <summary>
        /// Vrti nasledujci vrchol stromu poda zadanej cesty.
        /// Pouit pri prejden cez vetky rovnako nazvan vrcholy v intancii konceptu.
        /// Ak nenjde iadny al vyhovujci vrchol, vrti null.
        /// </summary>
        /// <param name="pathString">Zadan cesta bodkovm zpisom.</param>
        /// <param name="lastNode">Posledn njden vrchol, za ktorm bude hadanie pokraova.</param>
        /// <returns>Vrti nasledujci vrchol so zadanou cestou.</returns>
        public DNode GetNodeFromPath(String pathString, DNode lastNode) 
        {
            DNode found = GetNodeFromPath(pathString.Split('.'), lastNode);
            if (found == lastNode) return null;
            else return found;
        }

        /// <summary>
        /// Vrti vrchol poda zadanej cesty, ktor nasleduje za lastNode.
        /// Ak nenjde takto vrchol, vrti lastNode.
        /// </summary>
        /// <param name="path">Bodkov zpis rozdelen do poa reazcov.</param>
        /// <param name="lastNode">Posledn pouit vrchol s danou cestou.</param>
        /// <returns>Vrti vrchol s danou cestou, ktor sa nachdza a za danm lastNode.</returns>
        private DNode GetNodeFromPath(String[] path, DNode lastNode)
        {
            bool foundLastNode = (lastNode == null) ? true : false;
            if (path.Length == 1) {
                // vrati lastNode, ak ho nasiel a nenasiel dalsi vyhovujuci za nim
                // vrati najdeny vrchol, ak nasiel lastNode a aj ten vrchol
                // vrati null, ak nenasiel ani lastNode
                return (DNode)GetChild(path[0], (AiItemHolder)lastNode);
            } 
            else {
                // vytvor pole bez prveho prvku na vnorenie
                String[] cutpath = new String[path.Length - 1];
                Array.Copy(path, 1, cutpath, 0, path.Length - 1);
                DNode last = null;
                // prechadzaj vsetky moznosti
                while ((last = GetNode(path[0], last)) != null)
                {
                    DNode found = last.GetNodeFromPath(cutpath, (foundLastNode) ? null : lastNode);
                    // ak nieco nasiel, zisti ci to je lastNode
                    if (found != null) {
                        if (found == lastNode) foundLastNode = true; // ak je to lastNode, nastavime to
                        else return found; // inak sme nasli ten za nim a mozme vybublat (USPECH)
                    } 
                }
                // neuspesne vynorenie, ak sme nasli lastNode, dame to vediet aj nad sebou
                return (foundLastNode) ? lastNode : null;
            }
        }

        /// <summary>
        /// Njde potomka so zadanm menom na jednej rovni v rtane tch o patria pod nejak groupu.
        /// </summary>
        /// <param name="featureName">Meno hadanho potomka</param>
        /// <returns>Prv njden potomok s takm menom, null ak iaden tak neexistuje</returns>
        public DFeature GetChildFeature(string featureName)
        {
            return this.FindChildFeaturesTrasparently(featureName, this);
        }

        /// <summary>
        /// Rekurzvna metda implementujca njdenie potomka so zadanm menom na jednej 
        /// rovni v rtane tch o patria pod nejak groupu.
        /// </summary>
        /// <param name="featureName">Meno hadanho potomka</param>
        /// <param name="node">Uzol, z ktorho zaname hada</param>
        /// <returns>Prv njden potomok s takm menom, null ak iaden tak neexistuje</returns>
        private DFeature FindChildFeaturesTrasparently(string featureName, DNode node)
        {
            foreach (DNode child in node.Children)
            {
                if (child.GetType() == typeof(DGroup))
                {
                    DNode match = FindChildFeaturesTrasparently(featureName, child);
                    if (match != null)
                    {
                        return (DFeature)match;
                    }
                }
                else if (child.GetType() == typeof(DFeature))
                {
                    if (child.Name == featureName)
                    {
                        return (DFeature)child;
                    }
                }
            }

            return null;
        }

        /// <summary>
		/// Virtulna metda uren pre budce pouitie. Odvoden triedy DFeature a DGroup ju musia preai.
		/// Me sa poui len pre intancie.
		/// </summary>
		/// <returns>Pravda ak je uzol platn (validn) a aj vetci jeho potomkovia s platn, nepravda v opanom prpade.</returns>
		virtual public bool Validate()
		{
			return false;
		}

        /// <summary>
        /// Kpia vrcholu DiagramNode.
        /// </summary>
        /// <returns>Vrti kpiu vrcholu DiagramNode.</returns>
        public override TreeNode Clone()
        {
            return base.Clone();
        }

        /// <summary>
        /// Vrti aktulnu AiItem obsahujcu as viazania.
        /// </summary>
        public AiItem NodeBindingTime
        {
            get
            {
                if (this.Project.BindingTime == null)
                {
                    return null;
                }
                else
                {
                    return this.AiItems.GetAi(this.Project.BindingTime.Key);
                }
            }
        }
    }
}
