using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Collections;
using System.Drawing;

namespace FM
{
    /// <summary>
    /// Trieda TreeNode zastreuje cel stromov truktru projektu.
    /// </summary>
    [Serializable]
    public class TreeNode
    {
        /// <summary>
        /// Zoznam potomkov uzla.
        /// </summary>
        private ArrayList children;
        /// <summary>
        /// Rodi uzla.
        /// </summary>
        private TreeNode parent;
        /// <summary>
        /// Obsahuje meno uzla na lepiu identifikciu pre pouvatea.
        /// </summary>
        protected String name;

        /// <summary>
        /// Vytvorenie koreovho uzla bez mena.
        /// </summary>
        public TreeNode()
        {
            parent = null;
            name = "NoNameRoot";
            children = new ArrayList();
        }

        /// <summary>
        /// Vytvorenie koreovho uzla s menom.
        /// </summary>
        /// <param name="getName"></param>
        public TreeNode(String getName)
        {
            parent = null;
            name = getName;
            children = new ArrayList();
        }

        /// <summary>
        /// Kontruktor zabezpe vytvorenie uzla a naviazanie uzla na svojho rodia.
        /// </summary>
        /// <param name="getParent">Rodi pre uzol</param>
        public TreeNode(TreeNode getParent)
        {
            parent = getParent;
            name = "NoName";
            children = new ArrayList();

            if (parent != null)
            {
                parent.AddChild(this);
            }
        }

        /// <summary>
        /// Vytvorenie uzla s urenm rodiom a menom.
        /// </summary>
        /// <param name="getParent">Rodi, na ktorho sa naviae vytvran uzol</param>
        /// <param name="getName">Meno pre vytvran uzol</param>
        public TreeNode(TreeNode getParent, String getName)
        {
            parent = getParent;
            name = getName;
            children = new ArrayList();

            if (parent != null)
            {
                parent.AddChild(this);
            }
        }

        /// <summary>
        /// Zoznam potomkov uzla.
        /// </summary>
        public ArrayList Children
        {
            get { return children; }
        }

        /// <summary>
        /// Zoznam susedov uzla.
        /// </summary>
        public ArrayList GetNeighbours()
        {
            ArrayList neighbours = new ArrayList();
            if (parent != null)
            {
                foreach (TreeNode item in parent.Children)
                {
                    neighbours.Add(item);
                }
                neighbours.Remove(this);
            }
            return neighbours;
        }

        /// <summary>
        /// Rodi uzla.
        /// </summary>
        public TreeNode Parent
        {
            get { return parent; }
        }

        /// <summary>
        /// Meno uzla.
        /// </summary>
        virtual public String Name
        {
            get { return name; }
        }

        /// <summary>
        /// Vpis detailov uzla.
        /// </summary>
        protected virtual String Dump()
        {
            return "Name: " + Name + "\r\n";
        }

        /// <summary>
        /// Poet potomkov uzla.
        /// </summary>
        /// <returns></returns>
        public int CountChildren()
        {
            return children.Count;
        }
        /// <summary>
        /// Poet susedov uzla.
        /// </summary>
        /// <returns></returns>
        public int CountNeighbours()
        {
            if (Parent != null) return (Parent.CountChildren() - 1);
            else return 0;
        }

        /// <summary>
        /// Pridanie novho potomka rodiovi. Potomkovi nastav novho rodia. Pouvatesk funkcia.
        /// </summary>
        /// <param name="item">Uzol pridvan ako potomok.</param>
        public void AddChild(TreeNode item)
        {
            item.SetParent(this);
            if (!children.Contains(item)) children.Add(item);
        }

        /// <summary>
        /// Pridanie novho potomka rodiovi za zvolenho potomka. 
        /// Umouje pridvanie novch vlastnost na sprvnu pozciu v strome.
        /// Potomkovi nastav novho rodia. 
        /// Pouvatesk funkcia.
        /// </summary>
        /// <param name="item">Uzol pridvan ako potomok.</param>
        /// <param name="at">Uzol, za ktor bude pridan potomok. Ak je null, zarad sa potomok na prv miesto.</param>
        public void AddChildAfter(TreeNode item, TreeNode at)
        {
            item.SetParent(this);
            if (!children.Contains(item)) {
                if (children.Contains(at)) // ak sme nali takto vrchol v strome, vlome nov vrchol za ho
                    children.Insert(children.IndexOf(at) + 1, item); 
                else children.Insert(0, item); // ak sme takto vrchol nenali, vlome vrchol na zaiatok zoznamu
            } 
        }

        /// <summary>
        /// Odstrni potomka z uzla a z potomka vytvor koreov uzol. Pouvatesk funkcia.
        /// </summary>
        /// <param name="item">Uzol ktor bude odstrnen</param>
        public void RemoveChild(TreeNode item)
        {
            foreach (TreeNode child in children)
            if (children.Contains(item)) item.parent = null;
            if (children.Contains(item)) children.Remove(item);
        }

        /// <summary>
        /// Metda vrti zoznam uzlov ako cestu ku koreovmu uzlu.
        /// </summary>
        /// <returns></returns>
        public ArrayList GetNodeRouteToRoot()
        {
            TreeNode currentParent = parent;
            ArrayList routeToRoot = new ArrayList();
            while (currentParent != null)
            {
                routeToRoot.Add(currentParent);
                currentParent = currentParent.Parent;
            }
            return routeToRoot;
        }

        /// <summary>
        /// Vrti cestu v bodkovom zpise od vrcholu stromu.
        /// </summary>
        /// <returns>Cesta od vrcholu stromu v bodkovom zpise.</returns>
        public string GetNodePathFromRoot()
        {
            ArrayList routeToRoot = this.GetNodeRouteToRoot();
            routeToRoot.Reverse();

            string pathFromRoot = "";
            foreach (TreeNode node in routeToRoot)
            {
                pathFromRoot += node.Name;
                pathFromRoot += ".";
            }

            pathFromRoot += this.Name;

            return pathFromRoot;
        }

        /// <summary>
        /// Vytvor textov vstup, ktor reprezentuje stromov truktru tohto uzla a jeho potomkov.
        /// </summary>
        /// <returns>Obsahuje textov vstup so stromovou truktrou.</returns>
        public String ShowTree()
        {
            return ShowTree(0);
        }

        /// <summary>
        /// Rekurzvna as pre vytvorenie stromovej truktry.
        /// </summary>
        /// <param name="depth">Hbka.</param>
        /// <returns>Textov vstup so stromovou truktrou.</returns>
        private String ShowTree(int depth)
        {
            String output = "";
            for (int i = 0; i < depth; i++) output += "| ";
            output += "+ ";
            output += Dump() + "\r\n";
            foreach (TreeNode item in children)
                output += item.ShowTree(depth + 1);
            return output;
        }
        
        /// <summary>
        /// Zobraz truktru v ovldacom prvku typu TreeView, ktor je predan ako parameter.
        /// Metda zabezpeuje vetky potrebn nleitosti, vrtane rozbalenia celho stromu.
        /// </summary>
        /// <param name="control">Odkaz na ovldac prvok typu TreeView, v ktorom sa m zobrazi truktra</param>
        public void ShowTreeControl(TreeView control)
        {
            string pathSelectedNode;
            if (control.SelectedNode != null)
            {
                DNode selectedNode = (DNode)control.SelectedNode.Tag;
                pathSelectedNode = selectedNode.GetNodePathFromRoot();
            }
            else
            {
                pathSelectedNode = "";
            }

            control.BeginUpdate();
            control.Nodes.Clear();

            ShowTreeControl(control.Nodes);

            control.CheckBoxes = true;
            control.EndUpdate();
//            control.ExpandAll();

            System.Windows.Forms.TreeNode selectNode = GetSelectNodeInTreeView(control, pathSelectedNode);
            control.SelectedNode = selectNode;
        }

        /// <summary>
        /// Vrti uzol zobrazovacieho prvku TreeView poda zadanej hiererchickej cesty.
        /// </summary>
        /// <param name="control">TreeView zobrazovac prvok</param>
        /// <param name="pathSelectedNode">V Bodkovej notci zapsan cesta</param>
        /// <returns></returns>
        public System.Windows.Forms.TreeNode GetSelectNodeInTreeView(TreeView control, string pathSelectedNode)
        {
            string[] pathArray = pathSelectedNode.Split('.');

            System.Windows.Forms.TreeNode foundNode = null;

            if (control.TopNode.Text == pathArray[0])
            {
                foundNode = control.TopNode;
            }
            else
            {
                return null;
            }

            for (int a = 1; a < pathArray.Length; a++)
            {
                System.Windows.Forms.TreeNode actNode = foundNode.FirstNode;

                while (actNode != null)
                {
                    if (actNode.Text == pathArray[a])
                    {
                        break;
                    }
                    actNode = actNode.NextNode;
                }

                if (actNode == null)
                {
                    return null;
                }
                else
                {
                    foundNode = actNode;
                }
            }

            return foundNode;
        }

        /// <summary>
        /// Rekurzvne vytvor vntorn reprezentciu stromovej truktry potrebn pre jej
        /// zobrazenie v ovldacom prvku typu TreeView.
        /// </summary>
        /// <param name="nodes">Kolekcia uzlov na konkrtnej rovni truktry, kam sa vlo al uzol.</param>
        public void ShowTreeControl(TreeNodeCollection nodes)
        {
            System.Windows.Forms.TreeNode newNode = nodes.Add(Name);
            newNode.Name = Dump();
            newNode.Tag = this;
            newNode.EnsureVisible();

            if (this.GetType() == typeof(DFeature) || this.GetType() == typeof(DGroup))
            {
                if (((CardinalityHolder)this).Diagram.IsInstance == true)
                {
                    newNode.ForeColor = Color.Blue;

                    if (this.GetType() == typeof(DFeature))
                    {
						if (((DFeature)this).ReadOnly)
						{
							newNode.ForeColor = Color.DarkGray;
						}
						if (((DFeature)this).Unavailable)
						{
							newNode.ForeColor = Color.Gray;
						}
						newNode.Checked = ((DFeature)this).Selected;
                    }
                }
            }

            foreach (TreeNode item in children)
            {
                item.ShowTreeControl(newNode.Nodes);
            }
        }

        /// <summary>
        /// Prirad uzlu inho rodia. Zrove uzol odstrni s pvodnho stromu. Je to private funkcia.
        /// Poia potrebujete zmeni rodia, urobte to takto: novyRodic.AddChild(novyPotomokKtoremuSaZmeniRodic);
        /// </summary>
        /// <param name="item">Nov rodi.</param>
        private void SetParent(TreeNode item)
        {
            if ((parent != null)&&(parent != item)) parent.RemoveChild(this);
            parent = item;
        }

        /// <summary>
        /// Vytvra kpiu objektu. Bez rodia a bez potomkov.
        /// </summary>
        /// <returns>Vrti koprovan vrchol.</returns>
        public virtual TreeNode Clone()
        {
            TreeNode copy = (TreeNode)this.MemberwiseClone();
            copy.parent = null;
//            copy.name = Name + "Clone";
            copy.children = new ArrayList();

            return copy;
        }

        /// <summary>
        /// Vytvra kpiu objektu a jeho potomkov. Kopruje cel strom.
        /// </summary>
        /// <returns>Vrti koreov vrchol koprovanho stromu.</returns>
        public TreeNode CloneTree()
        {
            TreeNode copy = this.Clone(); // Mal by najst tu spravnu Clone() metodu cez virtual/override

            foreach (TreeNode item in children)
            {
                copy.AddChild(item.CloneTree());
            }
            return copy;
        }

        /// <summary>
        /// Verejn Dump, kee Name bolo niekde prepsan a neobsahuje Dump(),
        /// bolo potrebn doasne sprstupni Dump().
        /// </summary>
        /// <returns>Vrti vpis obsahu vlastnosti.</returns>
        public string DumpPublic()
        {
            return Dump();
        }
    }
}
